#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <time.h>

typedef struct {
    int tid;
    void (*lock)();
    void (*unlock)();
    void* lockval;
    unsigned long int* counter;
    unsigned long int num_incs;
    char* lockname;
} thread_args_t; 

void lock_tas_asm(int* lock) {

    register unsigned char _res = 1;
    while (_res != 0) {
     __asm__ __volatile__(
        "xchg %0,%1"
       :"+q" (_res), "+m" (*lock)
       : 
       : );
    }

}

void lock_tatas_asm(int* lock) {

    register unsigned char _res = 1;
    while (*lock == 1) { }
    while (_res != 0) {
         __asm__ __volatile__(
        "xchg %0,%1"
        :"+q" (_res), "+m" (*lock)
        : 
        : );
    }

}

void lock_tas_expbo_asm(int* lock) {

    int time = 1;
    register unsigned char _res = 1;
    while (_res != 0) {
         __asm__ __volatile__(
        "xchg %0,%1"
        :"+q" (_res), "+m" (*lock)
        : 
        : "memory");
        if (_res == 1) {
            time *= 2;
            for (int i=0; i<time; i++) {
                __asm__ __volatile__("nop");
            }
        }
    }

}

void unlock_tas_asm(int* lock) {
    __asm__ __volatile__(
        "mfence" 
       :
       :
       :"memory");
    *lock = 0;
}

void unlock_tatas_asm(int* lock) {
    __asm__ __volatile__(
        "mfence" 
       :
       :
       :"memory");
    *lock = 0;
}

void unlock_tas_expbo_asm(int* lock) {
    __asm__ __volatile__(
        "mfence" 
       :
       :
       :"memory");
    *lock = 0;
}

void init_tas_asm(thread_args_t* ta, int threads, int num_incs) {

    int* lockval = malloc(sizeof(int));
    unsigned long int* counter = malloc(sizeof(unsigned long int));
    *counter = 0;
    *lockval = 0;
    for (int t=0; t<threads; t++) {
        ta[t].tid = t;
        ta[t].num_incs = num_incs;
        ta[t].lockval = lockval;
        ta[t].counter = counter;
        ta[t].lock = lock_tas_asm;
        ta[t].unlock = unlock_tas_asm;
        ta[t].lockname = "TAS_ASM";
    }

}

void init_tatas_asm(thread_args_t* ta, int threads, int num_incs) {

    int* lockval = malloc(sizeof(int));
    unsigned long int* counter = malloc(sizeof(unsigned long int));
    *counter = 0;
    *lockval = 0;
    for (int t=0; t<threads; t++) {
        ta[t].tid = t;
        ta[t].num_incs = num_incs;
        ta[t].lockval = lockval;
        ta[t].counter = counter;
        ta[t].lock = lock_tatas_asm;
        ta[t].unlock = unlock_tatas_asm;
        ta[t].lockname = "TATAS_ASM";
    }

}

void init_tas_expbo_asm(thread_args_t* ta, int threads, int num_incs) {

    int* lockval = malloc(sizeof(int));
    unsigned long int* counter = malloc(sizeof(unsigned long int));
    *counter = 0;
    *lockval = 0;
    for (int t=0; t<threads; t++) {
        ta[t].tid = t;
        ta[t].num_incs = num_incs;
        ta[t].lockval = lockval;
        ta[t].counter = counter;
        ta[t].lock = lock_tas_expbo_asm;
        ta[t].unlock = unlock_tas_expbo_asm;
        ta[t].lockname = "TAS_EXPBO_ASM";
    }

}

void finalize_tas_asm(thread_args_t* ta, int threads) {

        free(ta[0].counter);
        free(ta[0].lockval);

}

void finalize_tatas_asm(thread_args_t* ta, int threads) {

        free(ta[0].counter);
        free(ta[0].lockval);

}

void finalize_tas_expbo_asm(thread_args_t* ta, int threads) {

        free(ta[0].counter);
        free(ta[0].lockval);

}

void* test_lock(void* args) {

    thread_args_t* ta = args;

    for (int i=0; i<ta->num_incs; i++) { 
        ta->lock(ta->lockval);
        *(ta->counter) += 1;
        ta->unlock(ta->lockval);
    }
}

double timediff(struct timespec* start, struct timespec* stop) {
    double stoptime = stop->tv_nsec + stop->tv_sec * 1e9;
    double starttime = start->tv_nsec + start->tv_sec * 1e9;
    return stoptime - starttime;
}

void perform_test(void (*init_func)(), void (*finalize_func)(),
                  int incs_per_thread, int max_threads) {

    static int firstprint = 1; // so that we know when to print header

    for (int num_threads=1; num_threads<max_threads; num_threads++) {
    
        // alocate space for thread management datastructure
        pthread_t* threads = malloc(num_threads*sizeof(pthread_t));
        assert(threads != NULL);

        // allocate space for thread arguments
        thread_args_t* thread_args = malloc(num_threads*sizeof(thread_args_t));
        assert(thread_args != NULL);
        init_func(thread_args, num_threads, incs_per_thread);

        // threads should be joinable
        pthread_attr_t attr;
        pthread_attr_init(&attr);
        pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
        
        struct timespec time1, time2;
        clock_gettime(CLOCK_REALTIME, &time1);

        // create num_thread threds
        for(int t=0; t<num_threads; t++){
            int rc = pthread_create(&threads[t], &attr, test_lock, 
                                (void *) &thread_args[t]);
            assert(rc == 0);
        }

        // wait for threads to complete
        pthread_attr_destroy(&attr);
        for(int t=0; t<num_threads; t++) {
            int rc = pthread_join(threads[t], NULL);
            assert(rc == 0);
        }

        clock_gettime(CLOCK_REALTIME, &time2);

        unsigned long int incs = 0;
        for (int t=0; t<num_threads; t++) incs += thread_args[t].num_incs;
        double time_per_update = timediff(&time1, &time2) / incs;
        char* correctness = "yes";
        if (*(thread_args[0].counter) != incs) correctness = "no";
        if (firstprint) {
            firstprint = 0;
            printf("%10s %20s %20s %12s\n", "threads", "lock", "ns_per_update", "correctness");
        }
        printf("%10i %20s %20.6lf %12s\n", num_threads, thread_args[0].lockname,
               time_per_update, correctness);

        finalize_func(thread_args, num_threads);
        free(threads);
        free(thread_args);
    }

}

int main(int argc, char** argv) {

    perform_test(init_tas_asm, finalize_tas_asm, 1e6, 8);
    perform_test(init_tatas_asm, finalize_tatas_asm, 1e6, 8);
    perform_test(init_tas_expbo_asm, finalize_tas_expbo_asm, 1e6, 8);
    pthread_exit(NULL);

}

